unit vstack;
{$IFDEF FPC}
  {$MODE DELPHI}{$H+}
{$ENDIF}

(***************************************)
(* Copyright (C) 2003, SHIRAISHI Kazuo *)
(***************************************)


interface

function getmemory(size:integer):pointer;
function GetZeroMemory(size:integer):pointer;
procedure freememory(size:integer);
procedure InitMemory;
procedure InitVirtualStack(size:Cardinal);

implementation
uses
 {$IFDEF win32}Windows,{$ENDIF}
 {$IFDEF Unix} BaseUnix, Unix, UnixType,{$ENDIF}
   base;

var
  StackBase :   pointer;
  StackBottom : pointer;
  StackLimit :  pointer;
  MaxStackSize :   Cardinal;
  StackSize :      Cardinal = 0;

procedure clear(var a; n:LongInt);assembler;
asm
   push   edi
   mov    edi,a
   mov    ecx,n
   xor    eax,eax
   rep    stosd
   pop    edi
end;



{$IFDEF Win32}

const StackDelta  =  $10000; // 64Kbytes

function intraise(i:Cardinal):Cardinal;
begin
    if i mod StackDelta =0 then
      result:=i
    else
      result:=((i div StackDelta)+1)*StackDelta
end;

procedure enlarge(size:integer);
var
    temp:Cardinal;
    temp1:pointer;
begin
    temp:=StackSize+intraise(size);
    temp1:=VirtualAlloc(StackBase,temp,
                                       MEM_COMMIT,PAGE_READWRITE);
    if temp1=nil then
          begin
             temp1:=VirtualAlloc(StackBase,StackSize,
                                       MEM_COMMIT,PAGE_READWRITE);
             setexception(VirtualStackOverflow);
          end;
    StackSize:=temp;
    StackBase:=temp1;
    Cardinal(StackLimit):=Cardinal(StackBase)+StackSize;
end;

function getmemory(size:integer):pointer;
begin
  if LongInt(StackLimit)-LongInt(StackBottom)<size then
     enlarge(size);
  GetMemory:=StackBottom;
  Inc(Longint(StackBottom),size);
end;

function GetZeroMemory(size:integer):pointer;
begin
  if LongInt(StackLimit)-LongInt(StackBottom)<size then
     enlarge(size);
  clear(StackBottom^, size div sizeof(longint));
  GetZeroMemory:=StackBottom;
  Inc(Longint(StackBottom),size);
end;

procedure freememory(size:integer);
begin
  Dec(Longint(StackBottom),size);
  if StackBottom<StackBase then  setexception(-107)
end;

procedure InitMemory;
begin
  StackBottom:=StackBase
end;

procedure InitVirtualStack(size:Cardinal);
begin
  MaxStackSize:=size;
  StackBase:=VirtualAlloc(nil,MaxStackSize,MEM_RESERVE,PAGE_READWRITE);
  while StackBase=nil do
    begin
      Dec(Cardinal(MaxStackSize),$1000000);
      StackBase:=VirtualAlloc(nil,MaxStackSize,MEM_RESERVE,PAGE_READWRITE);
    end;
  StackBottom:=StackBase;
  StackLimit:=StackBase;
end;

initialization


finalization

 VirtualFree(StackBase,MaxStackSize,MEM_DECOMMIT);
 VirtualFree(StackBase,0,MEM_RELEASE);


{$ELSE}


function getmemory(size:integer):pointer;
begin
  if LongInt(StackLimit)-LongInt(StackBottom)<size then
                   setexception(VirtualStackOverflow);
  GetMemory:=StackBottom;
  Inc(Longint(StackBottom),size);
end;

function GetZeroMemory(size:integer):pointer;
begin
  if LongInt(StackLimit)-LongInt(StackBottom)<size then
                   setexception(VirtualStackOverflow);
  clear(StackBottom^, size div sizeof(longint));
  GetZeroMemory:=StackBottom;
  Inc(Longint(StackBottom),size);
end;


procedure freememory(size:integer);
begin
  Dec(Longint(StackBottom),size);
end;

procedure InitMemory;
begin
  StackBottom:=StackBase
end;



procedure InitializeStackBase;
{$IFDEF Unix}
begin
 StackBase:=fpmmap(nil, StackSize, PROT_READ or PROT_WRITE
                  , MAP_PRIVATE or MAP_ANONYMOUS, 0, 0);
 while StackBase=MAP_FAILED do
    begin
       dec(StackSize, $4000000{64MB});
       StackBase:=fpmmap(nil, StackSize, PROT_READ or PROT_WRITE
                          , MAP_PRIVATE or MAP_ANONYMOUS, 0, 0);
    end;
end;
{$ELSE}
var flag:boolean;
begin
 flag:=false;
 repeat
   try
     GetMem(StackBase,StackSize);
     flag:=true;
   except
     dec(StackSize,$8000000)
   end;
 until flag=true;
end;
{$ENDIF}

procedure InitVirtualStack(size:Cardinal);
begin
 stacksize:=size;
 InitializeStackBase;
 StackBottom:=StackBase;
 Cardinal(StackLimit):=Cardinal(StackBase)+StackSize;
end;


initialization


finalization
{$IFNDEF Unix}
 FreeMem(stackBase,StackSize);
{$ENDIF}

{$ENDIF}

end.
